@@ -49,6 +49,7 @@ Package::Package(PackageStructure *structure)
     : d(new PackagePrivate())
 {
     d->structure = structure;
+
     if (d->structure) {
         d->structure.data()->initPackage(this);
     }
@@ -72,6 +73,11 @@ Package &Package::operator=(const Package &rhs)
     return *this;
 }
 
+bool Package::hasValidStructure() const
+{
+    return d->structure;
+}
+
 bool Package::isValid() const
 {
     if (!d->structure) {
@@ -94,10 +100,8 @@ bool Package::isValid() const
     //even if it's a big nested loop, usually there is one prefix and one location
     //so shouldn't cause too much disk access
     QHashIterator<QByteArray, ContentStructure> it(d->contents);
-    QString p = d->path;
-    if (!d->tempRoot.isEmpty()) {
-        p = d->tempRoot;
-    }
+    const QString rootPath = d->tempRoot.isEmpty() ? d->path : d->tempRoot;
+
     while (it.hasNext()) {
         it.next();
         if (!it.value().required) {
@@ -107,7 +111,7 @@ bool Package::isValid() const
         bool failed = true;
         foreach (const QString &path, it.value().paths) {
             foreach (const QString &prefix, d->contentsPrefixPaths) {
-                if (QFile::exists(p + prefix + path)) {
+                if (QFile::exists(rootPath + prefix + path)) {
                     failed = false;
                     break;
                 }
@@ -185,6 +189,27 @@ void Package::setDefaultPackageRoot(const QString &packageRoot)
     }
 }
 
+void Package::setFallbackPackage(const Plasma::Package &package)
+{
+    if ((d->fallbackPackage && d->fallbackPackage->path() == package.path() && d->fallbackPackage->metadata() == package.metadata()) ||
+        //can't be fallback of itself
+        (package.path() == path() && package.metadata() == metadata()) ||
+        d->hasCycle(package)) {
+        return;
+    }
+
+    d->fallbackPackage = new Package(package);
+}
+
+Plasma::Package Package::fallbackPackage() const
+{
+    if (d->fallbackPackage) {
+        return (*d->fallbackPackage);
+    } else {
+        return Package();
+    }
+}
+
 QString Package::servicePrefix() const
 {
     return d->servicePrefix;
@@ -287,7 +312,7 @@ QString Package::filePath(const char *fileType, const QString &filename) const
 {
     if (!d->valid) {
         //qDebug() << "package is not valid";
-        return QString();
+        return d->fallbackFilePath(fileType, filename);
     }
 
     const QString discoveryKey(fileType + filename);
@@ -302,7 +327,7 @@ QString Package::filePath(const char *fileType, const QString &filename) const
         //qDebug()<<d->contents.keys();
         if (!d->contents.contains(fileType)) {
             //qDebug() << "package does not contain" << fileType << filename;
-            return QString();
+            return d->fallbackFilePath(fileType, filename);
         }
 
         paths = d->contents[fileType].paths;
@@ -310,7 +335,7 @@ QString Package::filePath(const char *fileType, const QString &filename) const
         if (paths.isEmpty()) {
             //qDebug() << "no matching path came of it, while looking for" << fileType << filename;
             d->discoveries.insert(discoveryKey, QString());
-            return QString();
+            return d->fallbackFilePath(fileType, filename);
         }
     } else {
         //when filetype is empty paths is always empty, so try with an empty string
@@ -353,7 +378,7 @@ QString Package::filePath(const char *fileType, const QString &filename) const
     }
 
     //qDebug() << fileType << filename << "does not exist in" << prefixes << "at root" << d->path;
-    return QString();
+    return d->fallbackFilePath(fileType, filename);
 }
 
 QStringList Package::entryList(const char *key) const
@@ -420,8 +445,8 @@ void Package::setPath(const QString &path)
         return;
     }
 
-    // our dptr is shared, and it is almost certainly doing to change.
-    // hold onto the old pointer just incase!
+    // our dptr is shared, and it is almost certainly going to change.
+    // hold onto the old pointer just in case it does not, however!
     QExplicitlySharedDataPointer<PackagePrivate> oldD(d);
     d.detach();
 
@@ -448,11 +473,13 @@ void Package::setPath(const QString &path)
     QStringList paths;
     if (QDir::isRelativePath(path)) {
         QString p;
+
         if (d->defaultPackageRoot.isEmpty()) {
             p = path % "/";
         } else {
             p = d->defaultPackageRoot % path % "/";
         }
+
         if (QDir::isRelativePath(p)) {
             paths << QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, p, QStandardPaths::LocateDirectory);
         } else {
@@ -461,6 +488,7 @@ void Package::setPath(const QString &path)
                 paths << p;
             }
         }
+
         //qDebug() << "paths:" << p << paths << d->defaultPackageRoot;
     } else {
         const QDir dir(path);
@@ -494,6 +522,7 @@ void Package::setPath(const QString &path)
         }
     }
 
+
     // if nothing did change, then we go back to the old dptr
     if (d->path == previousPath) {
         d = oldD;
@@ -505,6 +534,8 @@ void Package::setPath(const QString &path)
     delete d->metadata;
     d->metadata = 0;
 
+    QString fallback;
+
     // uh-oh, but we didn't end up with anything valid, so we sadly reset ourselves
     // to futility.
     if (!d->valid) {
@@ -585,8 +616,17 @@ void Package::addDirectoryDefinition(const char *key, const QString &path, const
 
     if (d->contents.contains(key)) {
         s = d->contents[key];
+        if (s.paths.contains(path) && s.directory == true
+#ifndef PLASMA_NO_PACKAGE_EXTRADATA
+            && s.name == name
+#endif
+        ) {
+            return;
+        }
     }
 
+    d.detach();
+
 #ifndef PLASMA_NO_PACKAGE_EXTRADATA
     if (!name.isEmpty()) {
         s.name = name;
@@ -607,8 +647,16 @@ void Package::addFileDefinition(const char *key, const QString &path, const QStr
 
     if (d->contents.contains(key)) {
         s = d->contents[key];
+        if (s.paths.contains(path) && s.directory == false
+#ifndef PLASMA_NO_PACKAGE_EXTRADATA
+            && s.name == name
+#endif
+        ) {
+            return;
+        }
     }
 
+    d.detach();
 #ifndef PLASMA_NO_PACKAGE_EXTRADATA
     if (!name.isEmpty()) {
         s.name = name;
@@ -625,7 +673,10 @@ void Package::addFileDefinition(const char *key, const QString &path, const QStr
 
 void Package::removeDefinition(const char *key)
 {
-    d->contents.remove(key);
+    if (d->contents.contains(key)) {
+        d.detach();
+        d->contents.remove(key);
+    }
 }
 
 void Package::setRequired(const char *key, bool required)
@@ -636,6 +687,8 @@ void Package::setRequired(const char *key, bool required)
     }
 
     d.detach();
+    // have to find the item again after detaching: d->contents is a different object now
+    it = d->contents.find(key);
     it.value().required = required;
 }
 
@@ -656,6 +709,8 @@ void Package::setMimeTypes(const char *key, QStringList mimeTypes)
     }
 
     d.detach();
+    // have to find the item again after detaching: d->contents is a different object now
+    it = d->contents.find(key);
     it.value().mimeTypes = mimeTypes;
 #endif
 }
@@ -734,6 +789,7 @@ KJob *Package::uninstall(const QString &packageName, const QString &packageRoot)
 PackagePrivate::PackagePrivate()
     : QSharedData(),
       servicePrefix("plasma-applet-"),
+      fallbackPackage(0),
       metadata(0),
       externalPaths(false),
       valid(false),
@@ -756,6 +812,7 @@ PackagePrivate::~PackagePrivate()
         dir.removeRecursively();
     }
     delete metadata;
+    delete fallbackPackage;
 }
 
 PackagePrivate &PackagePrivate::operator=(const PackagePrivate &rhs)
@@ -765,6 +822,11 @@ PackagePrivate &PackagePrivate::operator=(const PackagePrivate &rhs)
     }
 
     structure = rhs.structure;
+    if (rhs.fallbackPackage) {
+        fallbackPackage = new Package(*rhs.fallbackPackage);
+    } else {
+        fallbackPackage = 0;
+    }
     path = rhs.path;
     contentsPrefixPaths = rhs.contentsPrefixPaths;
     servicePrefix = rhs.servicePrefix;
@@ -844,4 +906,38 @@ void PackagePrivate::createPackageMetadata(const QString &path)
     metadata = new KPluginInfo(metadataPath);
 }
 
+QString PackagePrivate::fallbackFilePath(const char *key, const QString &filename) const
+{
+    //don't fallback if the package isn't valid and never fallback the metadata file
+    if (qstrcmp(key, "metadata") != 0 && fallbackPackage && fallbackPackage->isValid()) {
+        return fallbackPackage->filePath(key, filename);
+    } else {
+        return QString();
+    }
+}
+
+bool PackagePrivate::hasCycle(const Plasma::Package &package)
+{
+    if (!package.d->fallbackPackage) {
+        return false;
+    }
+
+    //This is the Floyd cycle detection algorithm
+    //http://en.wikipedia.org/wiki/Cycle_detection#Tortoise_and_hare
+    Plasma::Package *slowPackage = const_cast<Plasma::Package *>(&package);
+    Plasma::Package *fastPackage = const_cast<Plasma::Package *>(&package);
+
+    while (fastPackage && fastPackage->d->fallbackPackage) {
+        //consider two packages the same if they have the same metadata
+        if ((fastPackage->d->fallbackPackage->metadata().isValid() && fastPackage->d->fallbackPackage->metadata() == slowPackage->metadata()) ||
+            (fastPackage->d->fallbackPackage->d->fallbackPackage && fastPackage->d->fallbackPackage->d->fallbackPackage->metadata().isValid() && fastPackage->d->fallbackPackage->d->fallbackPackage->metadata() == slowPackage->metadata())) {
+            qWarning() << "Warning: the fallback chain of " << package.metadata().pluginName() << "contains a cyclical dependency.";
+            return true;
+        }
+        fastPackage = fastPackage->d->fallbackPackage->d->fallbackPackage;
+        slowPackage = slowPackage->d->fallbackPackage;
+    }
+    return false;
+}
+
 } // Namespace
