@@ -52,45 +52,41 @@
 #include "kdeclarative/configpropertymap.h"
 #include <packageurlinterceptor.h>
 
-ContainmentInterface::ContainmentInterface(DeclarativeAppletScript *parent)
-    : AppletInterface(parent),
+ContainmentInterface::ContainmentInterface(DeclarativeAppletScript *parent, const QVariantList &args)
+    : AppletInterface(parent, args),
       m_wallpaperInterface(0),
       m_activityInfo(0)
 {
+    m_containment = static_cast<Plasma::Containment *>(appletScript()->applet()->containment());
+
     setAcceptedMouseButtons(Qt::AllButtons);
 
-    connect(containment(), &Plasma::Containment::appletRemoved,
+    connect(m_containment.data(), &Plasma::Containment::appletRemoved,
             this, &ContainmentInterface::appletRemovedForward);
-    connect(containment(), &Plasma::Containment::appletAdded,
+    connect(m_containment.data(), &Plasma::Containment::appletAdded,
             this, &ContainmentInterface::appletAddedForward);
-    connect(containment(), &Plasma::Containment::activityChanged,
-            this, &ContainmentInterface::activityChanged);
-    connect(containment(), &Plasma::Containment::activityChanged,
-            [=]() {
-                delete m_activityInfo;
-                m_activityInfo = new KActivities::Info(containment()->activity(), this);
-                connect(m_activityInfo, &KActivities::Info::nameChanged,
-                        this, &ContainmentInterface::activityNameChanged);
-                emit activityNameChanged();
-            });
-    connect(containment(), &Plasma::Containment::wallpaperChanged,
-            this, &ContainmentInterface::loadWallpaper);
-    connect(containment(), &Plasma::Containment::containmentTypeChanged,
-            this, &ContainmentInterface::containmentTypeChanged);
-
-     if (containment()->corona()) {
-         connect(containment()->corona(), &Plasma::Corona::availableScreenRegionChanged,
-                 this, &ContainmentInterface::availableScreenRegionChanged);
-     }
 
     if (!m_appletInterfaces.isEmpty()) {
         emit appletsChanged();
     }
+
+    connect(QCoreApplication::instance(), &QCoreApplication::aboutToQuit,
+            [=]() {
+                if (!m_containment) {
+                    return;
+                }
+                disconnect(m_containment.data(), &Plasma::Containment::appletRemoved,
+                           this, &ContainmentInterface::appletRemovedForward);
+            });
 }
 
 void ContainmentInterface::init()
 {
-    m_activityInfo = new KActivities::Info(containment()->activity(), this);
+    if (qmlObject()->rootObject()) {
+        return;
+    }
+
+    m_activityInfo = new KActivities::Info(m_containment->activity(), this);
     connect(m_activityInfo, &KActivities::Info::nameChanged,
             this, &ContainmentInterface::activityNameChanged);
     emit activityNameChanged();
@@ -98,13 +94,12 @@ void ContainmentInterface::init()
     AppletInterface::init();
 
     //Create the ToolBox
-    Plasma::Containment *pc = containment();
-    if (pc) {
+    if (m_containment) {
         KConfigGroup defaults;
-        if (pc->containmentType() == Plasma::Types::DesktopContainment) {
-            defaults = KConfigGroup(KSharedConfig::openConfig(pc->corona()->package().filePath("defaults")), "Desktop");
-        } else if (pc->containmentType() == Plasma::Types::PanelContainment) {
-            defaults = KConfigGroup(KSharedConfig::openConfig(pc->corona()->package().filePath("defaults")), "Panel");
+        if (m_containment->containmentType() == Plasma::Types::DesktopContainment) {
+            defaults = KConfigGroup(KSharedConfig::openConfig(m_containment->corona()->package().filePath("defaults")), "Desktop");
+        } else if (m_containment->containmentType() == Plasma::Types::PanelContainment) {
+            defaults = KConfigGroup(KSharedConfig::openConfig(m_containment->corona()->package().filePath("defaults")), "Panel");
         }
 
         Plasma::Package pkg = Plasma::PluginLoader::self()->loadPackage("Plasma/Generic");
@@ -115,21 +110,19 @@ void ContainmentInterface::init()
             pkg.setPath("org.kde.desktoptoolbox");
         }
 
-        PackageUrlInterceptor *interceptor = dynamic_cast<PackageUrlInterceptor *>(qmlObject()->engine()->urlInterceptor());
+        PlasmaQuick::PackageUrlInterceptor *interceptor = dynamic_cast<PlasmaQuick::PackageUrlInterceptor *>(qmlObject()->engine()->urlInterceptor());
         if (interceptor) {
             interceptor->addAllowedPath(pkg.path());
         }
 
         if (pkg.isValid()) {
             QObject *containmentGraphicObject = qmlObject()->rootObject();
-            if (containmentGraphicObject) {
-                QVariantHash toolboxProperties;
-                toolboxProperties["parent"] = QVariant::fromValue(containmentGraphicObject);
-                QObject *toolBoxObject = qmlObject()->createObjectFromSource(QUrl::fromLocalFile(pkg.filePath("mainscript")), 0, toolboxProperties);
-                if (toolBoxObject) {
-                    containmentGraphicObject->setProperty("toolBox", QVariant::fromValue(toolBoxObject));
-                }
-                qDebug() << "Loaded toolbox package" << pkg.path();
+
+            QVariantHash toolboxProperties;
+            toolboxProperties["parent"] = QVariant::fromValue(this);
+            QObject *toolBoxObject = qmlObject()->createObjectFromSource(QUrl::fromLocalFile(pkg.filePath("mainscript")), 0, toolboxProperties);
+            if (toolBoxObject && containmentGraphicObject) {
+                containmentGraphicObject->setProperty("toolBox", QVariant::fromValue(toolBoxObject));
             }
         } else {
             qWarning() << "Could not load toolbox package." << pkg.path();
@@ -146,6 +139,32 @@ void ContainmentInterface::init()
         QQmlProperty prop(qmlObject()->rootObject(), "anchors.fill");
         prop.write(expr.evaluate());
     }
+
+    if (!m_containment->wallpaper().isEmpty()) {
+        loadWallpaper();
+    }
+
+    connect(m_containment.data(), &Plasma::Containment::activityChanged,
+            this, &ContainmentInterface::activityChanged);
+    connect(m_containment.data(), &Plasma::Containment::activityChanged,
+            [ = ]() {
+                delete m_activityInfo;
+                m_activityInfo = new KActivities::Info(m_containment->activity(), this);
+                connect(m_activityInfo, &KActivities::Info::nameChanged,
+                        this, &ContainmentInterface::activityNameChanged);
+                emit activityNameChanged();
+            });
+    connect(m_containment.data(), &Plasma::Containment::wallpaperChanged,
+            this, &ContainmentInterface::loadWallpaper);
+    connect(m_containment.data(), &Plasma::Containment::containmentTypeChanged,
+            this, &ContainmentInterface::containmentTypeChanged);
+
+    if (m_containment->corona()) {
+        connect(m_containment->corona(), &Plasma::Corona::availableScreenRegionChanged,
+                this, &ContainmentInterface::availableScreenRegionChanged);
+        connect(m_containment->corona(), &Plasma::Corona::availableScreenRectChanged,
+                this, &ContainmentInterface::availableScreenRectChanged);
+    }
 }
 
 QList <QObject *> ContainmentInterface::applets()
@@ -163,43 +182,46 @@ void ContainmentInterface::setContainmentType(Plasma::Types::ContainmentType typ
     appletScript()->setContainmentType(type);
 }
 
-void ContainmentInterface::lockWidgets(bool locked)
+QVariantList ContainmentInterface::availableScreenRegion() const
 {
-    containment()->setImmutability(!locked ? Plasma::Types::Mutable : Plasma::Types::UserImmutable);
-    emit immutableChanged();
-}
-
-QRectF ContainmentInterface::screenGeometry(int id) const
-{
-    QRectF rect;
-    if (containment()->corona()) {
-        rect = QRectF(containment()->corona()->screenGeometry(id));
-    }
-
-    return rect;
-}
-
-QVariantList ContainmentInterface::availableScreenRegion(int id) const
-{
-    QRegion reg;
-    if (containment()->corona()) {
-        reg = containment()->corona()->availableScreenRegion(id);
+    QRegion reg = QRect(0, 0, width(), height());
+    int screenId = screen();
+    if (screenId > -1 && m_containment->corona()) {
+        reg = m_containment->corona()->availableScreenRegion(screenId);
     }
 
     QVariantList regVal;
     foreach (QRect rect, reg.rects()) {
+        //make it relative
+        QRect geometry = m_containment->corona()->screenGeometry(screenId);
+        rect.moveTo(rect.topLeft() - geometry.topLeft());
         regVal << QVariant::fromValue(QRectF(rect));
     }
     return regVal;
 }
 
-Plasma::Applet *ContainmentInterface::addApplet(const QString &plugin, const QVariantList &args, const QPoint &pos)
+QRect ContainmentInterface::availableScreenRect() const
+{
+    QRect rect(0, 0, width(), height());
+
+    int screenId = screen();
+
+    if (screenId > -1 && m_containment->corona()) {
+        rect = m_containment->corona()->availableScreenRect(screenId);
+        //make it relative
+        QRect geometry = m_containment->corona()->screenGeometry(screenId);
+        rect.moveTo(rect.topLeft() - geometry.topLeft());
+    }
+
+    return rect;
+}
+
+Plasma::Applet *ContainmentInterface::createApplet(const QString &plugin, const QVariantList &args, const QPoint &pos)
 {
     //HACK
     //This is necessary to delay the appletAdded signal (of containmentInterface) AFTER the applet graphics object has been created
     blockSignals(true);
-    Plasma::Applet *applet = containment()->createApplet(plugin, args);
-
+    Plasma::Applet *applet = m_containment->createApplet(plugin, args);
 
     if (applet) {
         QObject *appletGraphicObject = applet->property("_plasma_graphicObject").value<QObject *>();
@@ -208,8 +230,9 @@ Plasma::Applet *ContainmentInterface::addApplet(const QString &plugin, const QVa
 
         emit appletAdded(appletGraphicObject, pos.x(), pos.y());
         emit appletsChanged();
-    } else
+    } else {
         blockSignals(false);
+    }
     return applet;
 }
 
@@ -225,6 +248,61 @@ void ContainmentInterface::setAppletArgs(Plasma::Applet *applet, const QString &
     }
 }
 
+QObject *ContainmentInterface::containmentAt(int x, int y)
+{
+    foreach (Plasma::Containment *c, m_containment->corona()->containments()) {
+        ContainmentInterface *contInterface = c->property("_plasma_graphicObject").value<ContainmentInterface *>();
+
+        if (contInterface && contInterface->isVisible()) {
+            QWindow *w = contInterface->window();
+            if (w && w->geometry().contains(QPoint(window()->x(), window()->y()) + QPoint(x, y))) {
+                return contInterface;
+            }
+        }
+    }
+    return 0;
+}
+
+void ContainmentInterface::addApplet(AppletInterface *applet, int x, int y)
+{
+    if (!applet || applet->applet()->containment() == m_containment) {
+        return;
+    }
+
+    blockSignals(true);
+    m_containment->addApplet(applet->applet());
+    blockSignals(false);
+    emit appletAdded(applet, x, y);
+}
+
+QPointF ContainmentInterface::mapFromApplet(AppletInterface *applet, int x, int y)
+{
+    if (!applet->window() || !window()) {
+        return QPointF();
+    }
+
+    //x,y in absolute screen coordinates of current view
+    QPointF pos = applet->mapToScene(QPointF(x, y));
+    pos = QPointF(pos + applet->window()->geometry().topLeft());
+    //return the coordinate in the relative view's coords
+    return pos - window()->geometry().topLeft();
+}
+
+QPointF ContainmentInterface::mapToApplet(AppletInterface *applet, int x, int y)
+{
+    if (!applet->window() || !window()) {
+        return QPointF();
+    }
+
+    //x,y in absolute screen coordinates of current view
+    QPointF pos(x, y);
+    pos = QPointF(pos + window()->geometry().topLeft());
+    //the coordinate in the relative view's coords
+    pos = pos - applet->window()->geometry().topLeft();
+    //make it relative to applet coords
+    return pos - applet->mapToScene(QPointF(0, 0));
+}
+
 void ContainmentInterface::processMimeData(QMimeData *mimeData, int x, int y)
 {
     if (!mimeData) {
@@ -240,7 +318,8 @@ void ContainmentInterface::processMimeData(QMimeData *mimeData, int x, int y)
         const QStringList appletNames = data.split('\n', QString::SkipEmptyParts);
         foreach (const QString &appletName, appletNames) {
             qDebug() << "adding" << appletName;
-            addApplet(appletName, QVariantList(), QPoint(x, y));
+
+            metaObject()->invokeMethod(this, "createApplet", Qt::QueuedConnection, Q_ARG(QString, appletName), Q_ARG(QVariantList, QVariantList()), Q_ARG(QPoint, QPoint(x, y)));
         }
     } else if (mimeData->hasUrls()) {
         //TODO: collect the mimetypes of available script engines and offer
@@ -261,16 +340,14 @@ void ContainmentInterface::processMimeData(QMimeData *mimeData, int x, int y)
             KIO::MimetypeJob *job = KIO::mimetype(url, flags);
             m_dropPoints[job] = QPoint(x, y);
 
-
             QObject::connect(job, SIGNAL(result(KJob*)), this, SLOT(dropJobResult(KJob*)));
             QObject::connect(job, SIGNAL(mimetype(KIO::Job*,QString)),
-                                this, SLOT(mimeTypeRetrieved(KIO::Job*,QString)));
+                             this, SLOT(mimeTypeRetrieved(KIO::Job*,QString)));
 
             QMenu *choices = new QMenu("Content dropped");
             choices->addAction(QIcon::fromTheme("process-working"), i18n("Fetching file type..."));
             choices->popup(window() ? window()->mapToGlobal(QPoint(x, y)) : QPoint(x, y));
 
-
             m_dropMenus[job] = choices;
 #endif
         }
@@ -322,7 +399,7 @@ void ContainmentInterface::processMimeData(QMimeData *mimeData, int x, int y)
 
         if (!selectedPlugin.isEmpty()) {
 
-            Plasma::Applet *applet = addApplet(selectedPlugin, QVariantList(), QPoint(x, y));
+            Plasma::Applet *applet = createApplet(selectedPlugin, QVariantList(), QPoint(x, y));
             setAppletArgs(applet, pluginFormats[selectedPlugin], mimeData->data(pluginFormats[selectedPlugin]));
 
         }
@@ -343,14 +420,8 @@ void ContainmentInterface::clearDataForMimeJob(KIO::Job *job)
 void ContainmentInterface::dropJobResult(KJob *job)
 {
 #ifndef PLASMA_NO_KIO
-    KIO::TransferJob* tjob = dynamic_cast<KIO::TransferJob*>(job);
-    if (!tjob) {
-        qDebug() << "job is not a KIO::TransferJob, won't handle the drop...";
-        clearDataForMimeJob(tjob);
-        return;
-    }
     if (job->error()) {
-        qDebug() << "ERROR" << tjob->error() << ' ' << tjob->errorString();
+        qDebug() << "ERROR" << job->error() << ' ' << job->errorString();
     }
     // We call mimetypeRetrieved since there might be other mechanisms
     // for finding suitable applets. Cleanup happens there as well.
@@ -362,7 +433,7 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
 {
 #ifndef PLASMA_NO_KIO
     qDebug() << "Mimetype Job returns." << mimetype;
-    KIO::TransferJob* tjob = dynamic_cast<KIO::TransferJob*>(job);
+    KIO::TransferJob *tjob = dynamic_cast<KIO::TransferJob *>(job);
     if (!tjob) {
         qDebug() << "job should be a TransferJob, but isn't";
         clearDataForMimeJob(job);
@@ -392,7 +463,6 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
             return;
         }
 
-
         qDebug() << "Creating menu for:" << mimetype  << posi;
 
         appletList << Plasma::PluginLoader::self()->listAppletInfoForMimeType(mimetype);
@@ -420,7 +490,7 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
                 actionsToApplets.insert(action, info.pluginName());
                 qDebug() << info.pluginName();
             }
-            actionsToApplets.insert(choices->addAction(i18n("Icon")), "org.kde.icon");
+            actionsToApplets.insert(choices->addAction(i18n("Icon")), "org.kde.plasma.icon");
 
             QHash<QAction *, QString> actionsToWallpapers;
             if (!wallpaperList.isEmpty())  {
@@ -460,7 +530,7 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
                         m_wallpaperInterface->setUrl(tjob->url());
                     }
                 } else {
-                    Plasma::Applet *applet = addApplet(actionsToApplets[choice], QVariantList(), posi);
+                    Plasma::Applet *applet = createApplet(actionsToApplets[choice], QVariantList(), posi);
                     setAppletArgs(applet, mimetype, tjob->url().toString());
                 }
 
@@ -469,7 +539,7 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
             }
         } else {
             // we can at least create an icon as a link to the URL
-            Plasma::Applet *applet = addApplet("org.kde.icon", QVariantList(), posi);
+            Plasma::Applet *applet = createApplet("org.kde.plasma.icon", QVariantList(), posi);
             setAppletArgs(applet, mimetype, tjob->url().toString());
         }
     }
@@ -478,8 +548,6 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
 #endif // PLASMA_NO_KIO
 }
 
-
-
 void ContainmentInterface::appletAddedForward(Plasma::Applet *applet)
 {
     if (!applet) {
@@ -487,29 +555,24 @@ void ContainmentInterface::appletAddedForward(Plasma::Applet *applet)
     }
 
     QObject *appletGraphicObject = applet->property("_plasma_graphicObject").value<QObject *>();
-    QObject *contGraphicObject = containment()->property("_plasma_graphicObject").value<QObject *>();
+    QObject *contGraphicObject = m_containment->property("_plasma_graphicObject").value<QObject *>();
 
-    qDebug() << "Applet added on containment:" << containment()->title() << contGraphicObject
-             << "Applet: " << applet << applet->title() << appletGraphicObject;
+//     qDebug() << "Applet added on containment:" << m_containment->title() << contGraphicObject
+//              << "Applet: " << applet << applet->title() << appletGraphicObject;
 
-    if (contGraphicObject && appletGraphicObject) {
+    if (!appletGraphicObject) {
+        appletGraphicObject = new AppletInterface(applet, QVariantList(), this);
+        applet->setProperty("_plasma_graphicObject", QVariant::fromValue(appletGraphicObject));
+        static_cast<AppletInterface *>(appletGraphicObject)->init();
+    }
+
+    if (contGraphicObject) {
         appletGraphicObject->setProperty("visible", false);
         appletGraphicObject->setProperty("parent", QVariant::fromValue(contGraphicObject));
-
-    //if an appletGraphicObject is not set, we have to display some error message
-    } else if (contGraphicObject) {
-        QObject *errorUi = qmlObject()->createObjectFromSource(QUrl::fromLocalFile(containment()->corona()->package().filePath("appleterror")));
-
-        if (errorUi) {
-            errorUi->setProperty("visible", false);
-            errorUi->setProperty("parent", QVariant::fromValue(contGraphicObject));
-            errorUi->setProperty("reason", applet->launchErrorMessage());
-            appletGraphicObject = errorUi;
-        }
     }
 
     m_appletInterfaces << appletGraphicObject;
-    emit appletAdded(appletGraphicObject, 0, 0);
+    emit appletAdded(appletGraphicObject, -1, -1);
     emit appletsChanged();
 }
 
@@ -523,12 +586,12 @@ void ContainmentInterface::appletRemovedForward(Plasma::Applet *applet)
 
 void ContainmentInterface::loadWallpaper()
 {
-    if (containment()->containmentType() != Plasma::Types::DesktopContainment &&
-        containment()->containmentType() != Plasma::Types::CustomContainment) {
+    if (m_containment->containmentType() != Plasma::Types::DesktopContainment &&
+            m_containment->containmentType() != Plasma::Types::CustomContainment) {
         return;
     }
 
-    if (!containment()->wallpaper().isEmpty()) {
+    if (!m_containment->wallpaper().isEmpty()) {
         delete m_wallpaperInterface;
 
         m_wallpaperInterface = new WallpaperInterface(this);
@@ -541,7 +604,7 @@ void ContainmentInterface::loadWallpaper()
         QQmlProperty prop(m_wallpaperInterface, "anchors.fill");
         prop.write(expr.evaluate());
 
-        containment()->setProperty("wallpaperGraphicsObject", QVariant::fromValue(m_wallpaperInterface));
+        m_containment->setProperty("wallpaperGraphicsObject", QVariant::fromValue(m_wallpaperInterface));
     } else {
         if (m_wallpaperInterface) {
             m_wallpaperInterface->deleteLater();
@@ -552,7 +615,7 @@ void ContainmentInterface::loadWallpaper()
 
 QString ContainmentInterface::activity() const
 {
-    return containment()->activity();
+    return m_containment->activity();
 }
 
 QString ContainmentInterface::activityName() const
@@ -563,53 +626,84 @@ QString ContainmentInterface::activityName() const
     return m_activityInfo->name();
 }
 
-QList<QObject*> ContainmentInterface::actions() const
+QList<QObject *> ContainmentInterface::actions() const
 {
     //FIXME: giving directly a QList<QAction*> crashes
 
+    QStringList actionOrder;
+    actionOrder << "add widgets" << "manage activities" << "remove" << "lock widgets" << "run associated application" << "configure";
+    QHash<QString, QAction *> orderedActions;
     //use a multimap to sort by action type
-    QMultiMap<int, QObject*> actions;
-    foreach (QAction *a, containment()->actions()->actions()) {
-        if (a->isEnabled()) {
-            actions.insert(a->data().toInt(), a);
-        }
-    }
-    foreach (QAction *a, containment()->corona()->actions()->actions()) {
-        if (a->isEnabled()) {
-            if (a->objectName() == QStringLiteral("lock widgets")) {
-                //It is up to the Containment to decide if the user is allowed or not
-                //to lock/unluck the widgets, so corona should not add one when there is none
-                //(user is not allow) and it shouldn't add another one when there is already
-                //one
-                continue;
+    QMultiMap<int, QObject *> actions;
+    int i = 0;
+    foreach (QAction *a, m_containment->actions()->actions()) {
+        if (!actionOrder.contains(a->objectName())) {
+            //FIXME QML visualizations don't support menus for now, *and* there is no way to
+            //distinguish them on QML side
+            if (!a->menu()) {
+                actions.insert(a->data().toInt()*100 + i, a);
+                ++i;
             }
-            actions.insert(a->data().toInt(), a);
+        } else {
+            orderedActions[a->objectName()] = a;
         }
     }
-    return actions.values();
+
+    i = 0;
+    foreach (QAction *a, m_containment->corona()->actions()->actions()) {
+        if (a->objectName() == QStringLiteral("lock widgets") || a->menu()) {
+            //It is up to the Containment to decide if the user is allowed or not
+            //to lock/unluck the widgets, so corona should not add one when there is none
+            //(user is not allow) and it shouldn't add another one when there is already
+            //one
+            continue;
+        }
+
+        if (!actionOrder.contains(a->objectName())) {
+            actions.insert(a->data().toInt()*100 + i, a);
+        } else {
+            orderedActions[a->objectName()] = a;
+        }
+        ++i;
+    }
+    QList<QObject *> actionList = actions.values();
+
+    foreach (const QString &name, actionOrder) {
+        QAction *a = orderedActions.value(name);
+        if (a && a->isVisible() && !a->menu()) {
+            actionList << a;
+        }
+        ++i;
+    }
+
+    return actionList;
 }
 
-
-
-
 //PROTECTED--------------------
 
-void ContainmentInterface::mousePressEvent(QMouseEvent *event)
-{
-    event->setAccepted(containment()->containmentActions().contains(Plasma::ContainmentActions::eventToString(event)));
-}
-
 void ContainmentInterface::mouseReleaseEvent(QMouseEvent *event)
 {
+    event->setAccepted(m_containment->containmentActions().contains(Plasma::ContainmentActions::eventToString(event)));
+}
+
+void ContainmentInterface::mousePressEvent(QMouseEvent *event)
+
+{
+    //even if the menu is executed synchronously, other events may be processed
+    //by the qml incubator when plasma is loading, so we need to guard there
+    if (m_contextMenu) {
+        m_contextMenu.data()->close();
+        return;
+    }
+
     const QString trigger = Plasma::ContainmentActions::eventToString(event);
-    Plasma::ContainmentActions *plugin = containment()->containmentActions().value(trigger);
+    Plasma::ContainmentActions *plugin = m_containment->containmentActions().value(trigger);
 
     if (!plugin || plugin->contextualActions().isEmpty()) {
         event->setAccepted(false);
         return;
     }
 
-
     //the plugin can be a single action or a context menu
     //Don't have an action list? execute as single action
     //and set the event position as action data
@@ -627,29 +721,39 @@ void ContainmentInterface::mouseReleaseEvent(QMouseEvent *event)
         if (AppletInterface *ai = qobject_cast<AppletInterface *>(appletObject)) {
             if (ai->contains(ai->mapFromItem(this, event->posF()))) {
                 applet = ai->applet();
+                emit ai->contextualActionsAboutToShow();
                 break;
             } else {
                 ai = 0;
             }
         }
     }
-    qDebug() << "Invoking menu for applet" << applet;
+    //qDebug() << "Invoking menu for applet" << applet;
 
     QMenu desktopMenu;
 
+    m_contextMenu = &desktopMenu;
+
     if (applet) {
         addAppletActions(desktopMenu, applet, event);
     } else {
+        emit contextualActionsAboutToShow();
         addContainmentActions(desktopMenu, event);
     }
+
+    //this is a workaround where Qt now creates the menu widget
+    //in .exec before oxygen can polish it and set the following attribute
+    desktopMenu.setAttribute(Qt::WA_TranslucentBackground);
+    //end workaround
+
     desktopMenu.exec(event->globalPos());
-    event->accept();
+    event->setAccepted(true);
 }
 
 void ContainmentInterface::wheelEvent(QWheelEvent *event)
 {
     const QString trigger = Plasma::ContainmentActions::eventToString(event);
-    Plasma::ContainmentActions *plugin = containment()->containmentActions().value(trigger);
+    Plasma::ContainmentActions *plugin = m_containment->containmentActions().value(trigger);
 
     if (plugin) {
         if (event->delta() < 0) {
@@ -662,8 +766,6 @@ void ContainmentInterface::wheelEvent(QWheelEvent *event)
     }
 }
 
-
-
 void ContainmentInterface::addAppletActions(QMenu &desktopMenu, Plasma::Applet *applet, QEvent *event)
 {
     foreach (QAction *action, applet->contextualActions()) {
@@ -673,18 +775,22 @@ void ContainmentInterface::addAppletActions(QMenu &desktopMenu, Plasma::Applet *
     }
 
     if (!applet->failedToLaunch()) {
-        QAction *configureApplet = applet->actions()->action("configure");
-        if (configureApplet && configureApplet->isEnabled()) {
-            desktopMenu.addAction(configureApplet);
-        }
-
         QAction *runAssociatedApplication = applet->actions()->action("run associated application");
         if (runAssociatedApplication && runAssociatedApplication->isEnabled()) {
             desktopMenu.addAction(runAssociatedApplication);
         }
+
+        QAction *configureApplet = applet->actions()->action("configure");
+        if (configureApplet && configureApplet->isEnabled()) {
+            desktopMenu.addAction(configureApplet);
+        }
+        QAction *appletAlternatives = applet->actions()->action("alternatives");
+        if (appletAlternatives && appletAlternatives->isEnabled()) {
+            desktopMenu.addAction(appletAlternatives);
+        }
     }
 
-    QMenu *containmentMenu = new QMenu(i18nc("%1 is the name of the containment", "%1 Options", containment()->title()), &desktopMenu);
+    QMenu *containmentMenu = new QMenu(i18nc("%1 is the name of the containment", "%1 Options", m_containment->title()), &desktopMenu);
     addContainmentActions(*containmentMenu, event);
 
     if (!containmentMenu->isEmpty()) {
@@ -712,7 +818,8 @@ void ContainmentInterface::addAppletActions(QMenu &desktopMenu, Plasma::Applet *
         }
     }
 
-    if (containment()->immutability() == Plasma::Types::Mutable) {
+    if (m_containment->immutability() == Plasma::Types::Mutable &&
+        (m_containment->containmentType() != Plasma::Types::PanelContainment || m_containment->isUserConfiguring())) {
         QAction *closeApplet = applet->actions()->action("remove");
         //qDebug() << "checking for removal" << closeApplet;
         if (closeApplet) {
@@ -728,46 +835,44 @@ void ContainmentInterface::addAppletActions(QMenu &desktopMenu, Plasma::Applet *
 
 void ContainmentInterface::addContainmentActions(QMenu &desktopMenu, QEvent *event)
 {
-    if (containment()->corona()->immutability() != Plasma::Types::Mutable &&
-        !KAuthorized::authorizeKAction("plasma/containment_actions")) {
+    if (m_containment->corona()->immutability() != Plasma::Types::Mutable &&
+            !KAuthorized::authorizeKAction("plasma/containment_actions")) {
         //qDebug() << "immutability";
         return;
     }
 
     //this is what ContainmentPrivate::prepareContainmentActions was
     const QString trigger = Plasma::ContainmentActions::eventToString(event);
-    Plasma::ContainmentActions *plugin = containment()->containmentActions().value(trigger);
+    Plasma::ContainmentActions *plugin = m_containment->containmentActions().value(trigger);
 
     if (!plugin) {
         return;
     }
 
-    if (plugin->containment() != containment()) {
-        plugin->setContainment(containment());
+    if (plugin->containment() != m_containment) {
+        plugin->setContainment(m_containment);
 
         // now configure it
-        KConfigGroup cfg(containment()->corona()->config(), "ActionPlugins");
-        cfg = KConfigGroup(&cfg, QString::number(containment()->containmentType()));
+        KConfigGroup cfg(m_containment->corona()->config(), "ActionPlugins");
+        cfg = KConfigGroup(&cfg, QString::number(m_containment->containmentType()));
         KConfigGroup pluginConfig = KConfigGroup(&cfg, trigger);
         plugin->restore(pluginConfig);
     }
 
-
-    QList<QAction*> actions = plugin->contextualActions();
+    QList<QAction *> actions = plugin->contextualActions();
 
     if (actions.isEmpty()) {
         //it probably didn't bother implementing the function. give the user a chance to set
         //a better plugin.  note that if the user sets no-plugin this won't happen...
-        if ((containment()->containmentType() != Plasma::Types::PanelContainment &&
-             containment()->containmentType() != Plasma::Types::CustomPanelContainment) &&
-            containment()->actions()->action("configure")) {
-            desktopMenu.addAction(containment()->actions()->action("configure"));
+        if ((m_containment->containmentType() != Plasma::Types::PanelContainment &&
+                m_containment->containmentType() != Plasma::Types::CustomPanelContainment) &&
+                m_containment->actions()->action("configure")) {
+            desktopMenu.addAction(m_containment->actions()->action("configure"));
         }
     } else {
         desktopMenu.addActions(actions);
     }
 
-
     return;
 }
 
